/* Copyright (C) 2005, 2006, 2007, 2008, 2009 Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package frontend.sjava;

import compbase.*;

/**
 * Marker: functionality to mark a unit/method
 * 
 * @author S. Frenz
 * @version 091005 adopted changed Expr
 *  version 091001 adopted changed memory interface
 *  version 090718 added support for resetting MF_WRITTEN modifier, adopted changed Expr
 *  version 090626 added support for K_SEPC and K_NSPC marker
 *  version 090616 adopted changed ExCall
 *  version 090207 added copyright notice
 *  version 081015 added support for K_PRCD and K_NOPT marker
 *  version 080706 relaxed checking of method attributes in noInlineMthdObj mode
 *  version 080629 added check to ensure static method for inline-mark in noInlineMthdObj mode
 *  version 080613 adopted hasEffect->effectType, added stopBlockCoding
 *  version 080104 added support for K_PROF and K_NPRF marker
 *  version 071001 added mistake hint for new DEFINE-MARKER
 *  version 070909 optimized signature of Expr.resolve
 *  version 070713 added support to mark method not skipable
 *  version 070628 adopted changes for "Expr extends TypeRef"
 *  version 070617 bugfixed usage of mthdResolver
 *  version 070527 added mistake hint for new DEFINE-MARKER
 *  version 070331 added mistake hint for invalid use of DEFINE-MARKER
 *  version 070114 reduced access level where possible
 *  version 070111 adapted change in printPos and compErr
 *  version 070107 cleaned up
 *  version 061231 added mark to avoid method inlining
 *  version 061225 added mark to force method inlining
 *  version 061222 added enterCodeAddr and dynamized access
 *  version 061203 optimized calls to printPos and compErr
 *  version 061129 static TypeRef object moved dynamically to Context
 *  version 061109 removed optdebug-mark
 *  version 060607 initial version
 */

public class Marker extends CtxBasedConfig {
  protected boolean resolve(Expr ex, Unit unitContext, Mthd mthdContext, boolean insideLoop, Context ctx) {
    ExCall call;
    String name;
    
    if (ex instanceof ExCall) {
      call=(ExCall)ex;
      name=call.id;
      if (name.equals("interrupt")) {
        if (!resolveNoParam(call, ctx)) return false;
        mthdContext.marker|=Marks.K_INTR;
        return true;
      }
      else if (name.equals("debug")) {
        if (!resolveNoParam(call, ctx)) return false;
        mthdContext.marker|=Marks.K_DEBG;
        return true;
      }
      else if (name.equals("ignore")) { //ignore ;)
        call.baseType=TypeRef.T_VOID;
        call.effectType=Expr.EF_NORM;
        return true;
      }
      else if (name.equals("inline")) {
        if (!resolveNoParam(call, ctx)) return false;
        if (ctx.noInlineMthdObj && (mthdContext.modifier&(Mthd.M_OVERLD|Mthd.MA_INTFMD))!=0) {
          ex.printPos(ctx, "can not force inlining of overloaded/interface method");
          return false;
        }
        mthdContext.marker|=Marks.K_FINL;
        return true;
      }
      else if (name.equals("noInline")) {
        if (!resolveNoParam(call, ctx)) return false;
        mthdContext.marker|=Marks.K_NINL;
        return true;
      }
      else if (name.equals("profile")) {
        if (!resolveNoParam(call, ctx)) return false;
        mthdContext.marker|=Marks.K_PROF;
        return true;
      }
      else if (name.equals("noProfile")) {
        if (!resolveNoParam(call, ctx)) return false;
        mthdContext.marker|=Marks.K_NPRF;
        return true;
      }
      else if (name.equals("stackExtreme")) {
        if (!resolveNoParam(call, ctx)) return false;
        mthdContext.marker|=Marks.K_SEPC;
        return true;
      }
      else if (name.equals("noStackExtreme")) {
        if (!resolveNoParam(call, ctx)) return false;
        mthdContext.marker|=Marks.K_NSPC;
        return true;
      }
      else if (name.equals("printCode")) {
        if (!resolveNoParam(call, ctx)) return false;
        mthdContext.marker|=Marks.K_PRCD;
        return true;
      }
      else if (name.equals("noOptimization")) {
        if (!resolveNoParam(call, ctx)) return false;
        mthdContext.marker|=Marks.K_NOPT;
        return true;
      }
      else if (name.equals("enterCodeAddr")) {
        return resolveEnterCodeAddr(call, unitContext, mthdContext, insideLoop, ctx);
      }
      else if (name.equals("genCode")) {
        if (!resolveNoParam(call, ctx)) return false;
        mthdContext.modifier|=Mthd.M_NDCODE;
        return true;
      }
      else if (name.equals("stopBlockCoding")) {
        if (!resolveNoParam(call, ctx)) return false;
        call.effectType=StExpr.EF_STOP;
        return true;
      }
      else if (name.equals("offset") || name.equals("count")
          || name.equals("inlineArray") || name.equals("head")
          || name.equals("windows")) {
        ex.printPos(ctx, "invalid use of DEFINE-marker");
        return false;
      }
      else if (name.equals("treatAsUnwritten")) {
        if (!resolveTreatAsUnwritten(call, unitContext, mthdContext, insideLoop, ctx)) return false;
        call.baseType=TypeRef.T_VOID;
        call.effectType=Expr.EF_NORM;
        return true;
      }
    }
    ex.printPos(ctx, "unknown MARKER-command");
    return false;
  }
  
  //internal resolve-methods
  private static boolean resolveNoParam(ExCall call, Context ctx) {
    if (call.par==null) {
      //everything ok, set type
      call.baseType=TypeRef.T_VOID;
      call.effectType=Expr.EF_NORM;
      return true;
    }
    call.printPos(ctx, "MARKER expects no parameter");
    return false;
  }

  private boolean resolveEnterCodeAddr(ExCall call, Unit unitContext, Mthd mthdContext, boolean insideLoop, Context ctx) {
    ExClssMthdName mthdResolver;
    FilledParam pa;
    Expr paEx;
    int addr;
    
    //resolve all parameters, check if all parameters are constant integers
    pa=call.par;
    if (pa==null) {
      call.printPos(ctx, "MARKER.enterCodeAddr needs address");
      return false;
    }
    paEx=pa.expr;
    if (!paEx.resolve(unitContext, mthdContext, true, insideLoop, null, ctx)) return false;
    if (!paEx.isIntType() || paEx.calcConstantType(ctx)!=StdTypes.T_INT) {
      call.printPos(ctx, "MARKER.enterCodeAddr needs constant int value as destination");
      return false;
    }
    addr=paEx.getConstIntValue(ctx);
    pa=pa.nextParam;
    if (pa==null || (paEx=pa.expr)==null || !(paEx instanceof ExStr)) {
      call.printPos(ctx, "MARKER.enterCodeAddr needs string containing class as second parameter");
      return false;
    }
    mthdResolver=new ExClssMthdName(null, null, false, -1, -1, -1);
    mthdResolver.unitName=((ExStr)paEx).value;
    pa=pa.nextParam;
    if (pa==null || (paEx=pa.expr)==null || !(paEx instanceof ExStr)) {
      call.printPos(ctx, "MARKER.enterCodeAddr needs second string containing method as third parameter");
      return false;
    }
    mthdResolver.mthdName=((ExStr)paEx).value;
    if (pa.nextParam!=null) {
      call.printPos(ctx, "MARKER.enterCodeAddr needs only three parameters");
      return false;
    }
    mthdResolver.fileID=pa.fileID;
    mthdResolver.line=pa.line;
    mthdResolver.col=pa.col;
    if (!mthdResolver.resolve(unitContext, mthdContext, true, insideLoop, null, ctx)) return false;
    //everything ok, enter object
    mthdResolver.destMthd.enterCodeRef(null, addr, ctx);
    //set type
    call.baseType=TypeRef.T_VOID;
    call.effectType=Expr.EF_NORM;
    return true;
  }

  private static boolean resolveTreatAsUnwritten(ExCall call, Unit unitContext, Mthd mthdContext, boolean insideLoop, Context ctx) {
    FilledParam pa;
    Expr paEx;
    
    if ((pa=call.par)==null || pa.nextParam!=null) {
      call.printPos(ctx, "MARKER.treatAsUnwritten needs one parameter");
      return false;
    }
    if (!(paEx=pa.expr).resolve(unitContext, mthdContext, false, insideLoop, null, ctx)) return false;
    if (paEx instanceof ExDeRef) paEx=((ExDeRef)paEx).ri;
    if (!(paEx instanceof ExVar)) {
      call.printPos(ctx, "MARKER.treatAsUnwritten needs variable name");
      return false;
    }
    ((ExVar)paEx).dest.modifier&=~(AccVar.MF_ISWRITTEN|AccVar.MF_MAYBEWRITTEN);
    return true;
  }
}
